/*******************************************************************************
 *    Rubik's Cube 3D game
 *     Copyright (C) 2011  Harshad Sahasrabudhe (harshad.sahasrabudhe@gmail.com)
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc.,
 *     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * 
 * Contributors:
 * Harshad Sahasrabudhe (harshad.sahasrabudhe@gmail.com) - Core and UI development
 * Vishwajit Gaikwad (vishrockon7@gmail.com) - Testing
 ******************************************************************************/
package rubik.core.rubikgroup;


import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Stack;

import rubik.core.RubikBlockID;
import rubik.core.RubikRotationList;

public class RubikDisjointCycle {

	public static final int BLOCKS = 2;
	public static final int SIDES = 4;

	public static final boolean c0 = false;
	public static final boolean c1 = true;

	public static final int blockArraySize = 26;
	public static final int sideArraySize = 6;

	int type;

	boolean[][] transformationMatrix = null;

	public RubikDisjointCycle(int type) {

		this.type = type;
		if(type == BLOCKS) {
			transformationMatrix = new boolean[blockArraySize][blockArraySize];
			for(int i=0; i<blockArraySize; i++) {
				for(int j=0; j<blockArraySize; j++) {
					if(i == j) {
						transformationMatrix[i][j] = c1;
					}
					else {
						transformationMatrix[i][j] = c0;
					}
				}
			}
		}
		else if(type == SIDES) {
			transformationMatrix = new boolean[sideArraySize][sideArraySize];
			for(int i=0; i<sideArraySize; i++) {
				for(int j=0; j<sideArraySize; j++) {
					if(i == j) {
						transformationMatrix[i][j] = c1;
					}
					else {
						transformationMatrix[i][j] = c0;
					}
				}
			}
		}
	}
	
	public int getType() {
		return type;
	}
	
	public boolean[][] getTransformationMatrix() {
		return transformationMatrix;
	}
	
	public int differenceWith(RubikDisjointCycle rdc) {
		if(rdc.getType() != type) {
			return -1;
		}
		boolean[][] rdcMat = rdc.getTransformationMatrix();
		int size = getSize();
		int ret = 0;
		for(int i=0; i<size; i++) {
			for(int j=0; j<size; j++) {
				if(transformationMatrix[i][j] == c1 || rdcMat[i][j] == c1) {
					if(rdcMat[i][j] != transformationMatrix[i][j]) {
						ret ++;
					}
					break;
				}
			}
		}
		return ret;
		
	}
	
	@Override
	public boolean equals(Object rdc) {
		
		if(rdc instanceof RubikDisjointCycle) {
			RubikDisjointCycle c = ((RubikDisjointCycle) rdc);
			if(c.getType() != type) {
				return false;
			}
			int size;
			if(type == BLOCKS) {
				size = blockArraySize;
			}
			else {
				size = sideArraySize;
			}
			for(int i=0; i<size; i++) {
				for(int j=0; j<size; j++) {
					if(transformationMatrix[i][j] != 
						c.getTransformationMatrix()[i][j]) {
						return false;
					}
				}
			}
			return true;
		}
		else {
			return false;
		}
	}
	
	@Override
	public Object clone() {
		RubikDisjointCycle ret = new RubikDisjointCycle(type);
		int size = getSize();
		for(int i=0; i<size; i++) {
			for(int j=0; j<size; j++) {
				ret.getTransformationMatrix()[i][j] = transformationMatrix[i][j];
			}
		}
		return ret;
	}
	
	public int getSize() {
		if(type == BLOCKS) {
			return blockArraySize;
		}
		else {
			return sideArraySize;
		}
	}

	/* Notation:
	 * 
	 * Blocks
	 *             ___ ___ ___
	 *            |21 |11 |20 |  3<-D
	 *            |_o_|_o_|_o_|     |
	 *            | 3 |16 | 2 |3E   v   
	 *     ___ ___|_o_|_o_|_o_|_5_ _5_
	 *    |21 | 3 | 3 |16 | 2 | 2 |20 |   5<-T->3
	 *    |_b_|_b_|_y_|_y_|_y_|_g_|_g_|
	 *    |12 |17 |17 | 8 |15 |15 |10 |   5<-G->3
	 *    |_b_|_b_|_y_|_y_|_y_|_g_|_g_|
	 *    |18 | 0 | 0 |14 | 1 | 1 |19 |   5<-B->3
	 *    |_b_|_b_|_y_|_y_|_y_|_g_|_g_|
	 *            | 0 |14 | 1 |
	 *            |_r_|_r_|_r_|
	 *            |18 | 9 |19 |
	 *            |_r_|_r_|_r_|
	 * 
	 *              5   5   5
	 *              ^   ^   ^
	 *              V   F   R
	 *              v   v   v
	 *              3   3   3
	 *             ___ ___ ___
	 *            | 4 |22 | 5 |5C
	 *         ___|_r_|_r_|_r_|_3_
	 *        | 4 | 4 |22 | 5 | 5 |   3<-B->5
	 *        |_b_|_w_|_w_|_w_|_g_|
	 *        |25 |25 |13 |23 |23 |   3<-G->5
	 *        |_b_|_w_|_w_|_w_|_g_|
	 *        | 7 | 7 |24 | 6 | 6 |   3<-T->5
	 *        |_b_|_w_|_w_|_w_|_g_|
	 *            | 7 |24 | 6 |
	 *            |_o_|_o_|_o_|
	 *            
	 *       
	 *Sides
	 *             ___
	 *            | 4 |5(E D C) 
	 *         ___|y+_|_3_ ___
	 *        | 5 | 0 | 3 | 1 |    3<-(B G T)->5
	 *        |x-_|z+_|x+_|z-_|
	 *            | 2 |
	 *            |y-_|
	 *              
	 *              5
	 *              ^
	 *           (T G B)
	 *              v
	 *              3
	 */

	public void applyCommands(RubikRotationList list) {
		Stack<Integer> rotationList = list.getList();
		ListIterator<Integer> iter = rotationList.listIterator();
		int rotationCode;
		List<Integer> temp = new LinkedList<Integer>();

		while(iter.hasNext()) {

			rotationCode = iter.next();

			if(type == BLOCKS) {
				switch(rotationCode) {
				case RubikRotationList.e3:
					temp.clear();
					temp.add(0); temp.add(1); temp.add(2); temp.add(3);
					refreshMatrix(temp);
					temp.clear();
					temp.add(14); temp.add(15); temp.add(16); temp.add(17);
					refreshMatrix(temp);
					break;
				case RubikRotationList.e5:
					temp.clear();
					temp.add(0); temp.add(3); temp.add(2); temp.add(1);
					refreshMatrix(temp);
					temp.clear();
					temp.add(14); temp.add(17); temp.add(16); temp.add(15);
					refreshMatrix(temp);
					break;
				case RubikRotationList.d3:
					temp.clear();
					temp.add(0); temp.add(3); temp.add(2); temp.add(1);
					refreshMatrix(temp);
					temp.clear();
					temp.add(14); temp.add(17); temp.add(16); temp.add(15);
					refreshMatrix(temp);
					break;
				case RubikRotationList.d5:
					temp.clear();
					temp.add(0); temp.add(3); temp.add(2); temp.add(1);
					refreshMatrix(temp);
					temp.clear();
					temp.add(14); temp.add(17); temp.add(16); temp.add(15);
					refreshMatrix(temp);
					break;
				case RubikRotationList.c3:
					temp.clear();
					temp.add(4); temp.add(5); temp.add(6); temp.add(7);
					refreshMatrix(temp);
					temp.clear();
					temp.add(22); temp.add(23); temp.add(24); temp.add(25);
					refreshMatrix(temp);
					break;
				case RubikRotationList.c5:
					temp.clear();
					temp.add(4); temp.add(7); temp.add(6); temp.add(5);
					refreshMatrix(temp);
					temp.clear();
					temp.add(22); temp.add(25); temp.add(24); temp.add(23);
					refreshMatrix(temp);
					break;
				case RubikRotationList.r3:
					temp.clear();
					temp.add(1); temp.add(5); temp.add(6); temp.add(2);
					refreshMatrix(temp);
					temp.clear();
					temp.add(15); temp.add(19); temp.add(23); temp.add(20);
					refreshMatrix(temp);
					break;
				case RubikRotationList.r5:
					temp.clear();
					temp.add(1); temp.add(2); temp.add(6); temp.add(5);
					refreshMatrix(temp);
					temp.clear();
					temp.add(15); temp.add(20); temp.add(23); temp.add(19);
					refreshMatrix(temp);
					break;
				case RubikRotationList.f3:
					temp.clear();
					temp.add(8); temp.add(9); temp.add(13); temp.add(11);
					refreshMatrix(temp);
					temp.clear();
					temp.add(14); temp.add(22); temp.add(24); temp.add(16);
					refreshMatrix(temp);
					break;
				case RubikRotationList.f5:
					temp.clear();
					temp.add(8); temp.add(11); temp.add(13); temp.add(9);
					refreshMatrix(temp);
					temp.clear();
					temp.add(14); temp.add(16); temp.add(24); temp.add(22);
					refreshMatrix(temp);
					break;
				case RubikRotationList.v3:
					temp.clear();
					temp.add(0); temp.add(4); temp.add(7); temp.add(3);
					refreshMatrix(temp);
					temp.clear();
					temp.add(17); temp.add(18); temp.add(25); temp.add(21);
					refreshMatrix(temp);
					break;
				case RubikRotationList.v5:
					temp.clear();
					temp.add(0); temp.add(3); temp.add(7); temp.add(4);
					refreshMatrix(temp);
					temp.clear();
					temp.add(17); temp.add(21); temp.add(25); temp.add(18);
					refreshMatrix(temp);
					break;
				case RubikRotationList.t3:
					temp.clear();
					temp.add(2); temp.add(6); temp.add(7); temp.add(3);
					refreshMatrix(temp);
					temp.clear();
					temp.add(16); temp.add(20); temp.add(21); temp.add(24);
					refreshMatrix(temp);
					break;
				case RubikRotationList.t5:
					temp.clear();
					temp.add(2); temp.add(6); temp.add(7); temp.add(3);
					refreshMatrix(temp);
					temp.clear();
					temp.add(16); temp.add(24); temp.add(21); temp.add(20);
					refreshMatrix(temp);
					break;
				case RubikRotationList.g3:
					temp.clear();
					temp.add(8); temp.add(10); temp.add(13); temp.add(12);
					refreshMatrix(temp);
					temp.clear();
					temp.add(15); temp.add(23); temp.add(25); temp.add(17);
					refreshMatrix(temp);
					break;
				case RubikRotationList.g5:
					temp.clear();
					temp.add(8); temp.add(12); temp.add(13); temp.add(10);
					refreshMatrix(temp);
					temp.clear();
					temp.add(15); temp.add(17); temp.add(25); temp.add(23);
					refreshMatrix(temp);
					break;
				case RubikRotationList.b3:
					temp.clear();
					temp.add(2); temp.add(3); temp.add(4); temp.add(5);
					refreshMatrix(temp);
					temp.clear();
					temp.add(14); temp.add(19); temp.add(22); temp.add(18);
					refreshMatrix(temp);
					break;
				case RubikRotationList.b5:
					temp.clear();
					temp.add(2); temp.add(3); temp.add(4); temp.add(5);
					refreshMatrix(temp);
					temp.clear();
					temp.add(14); temp.add(18); temp.add(22); temp.add(19);
					refreshMatrix(temp);
					break;

				}
			}
			else if(type == SIDES) {
				switch(rotationCode) {
				case RubikRotationList.e3:
				case RubikRotationList.d3:
				case RubikRotationList.c3:
					temp.clear();
					temp.add(2); temp.add(3); temp.add(4); temp.add(5);
					refreshMatrix(temp);
					break;
				case RubikRotationList.e5:
				case RubikRotationList.d5:
				case RubikRotationList.c5:
					temp.clear();
					temp.add(2); temp.add(5); temp.add(4); temp.add(3);
					refreshMatrix(temp);
					break;
				case RubikRotationList.r3:
				case RubikRotationList.f3:
				case RubikRotationList.v3:
					temp.clear();
					temp.add(0); temp.add(2); temp.add(1); temp.add(4);
					refreshMatrix(temp);
					break;
				case RubikRotationList.r5:
				case RubikRotationList.f5:
				case RubikRotationList.v5:
					temp.clear();
					temp.add(0); temp.add(4); temp.add(1); temp.add(2);
					refreshMatrix(temp);
					break;
				case RubikRotationList.t3:
				case RubikRotationList.g3:
				case RubikRotationList.b3:
					temp.clear();
					temp.add(0); temp.add(3); temp.add(1); temp.add(5);
					refreshMatrix(temp);
					break;
				case RubikRotationList.t5:
				case RubikRotationList.g5:
				case RubikRotationList.b5:
					temp.clear();
					temp.add(0); temp.add(5); temp.add(1); temp.add(3);
					refreshMatrix(temp);
					break;
				}				
			}
		}
	}
	private void refreshMatrix(List<Integer> l1) {
		
		int size = getSize();

		boolean[][] m1 = new boolean[size][size];
		boolean[][] m2 = new boolean[size][size];
		int next, cur;

		for(int i=0; i<size; i++) {
			for(int j=0; j<size; j++) {
				if(i == j) {
					m1[i][j] = c1;
				}
				else {
					m2[i][j] = c0;
				}
				m2[i][j] = transformationMatrix[i][j];
			}
		}

		for(int i=0; i<l1.size(); i++) {
			cur = l1.get(i);
			if(i == l1.size()-1) {
				next = l1.get(0);
			}
			else {
				next = l1.get(i+1);
			}
			m1[cur][next] = c1;
			m1[cur][cur] = c0;
		}

		for(int i=0; i<size; i++) {
			for(int j=0; j<size; j++) {
				for(int k=0; k<size; k++) {

					if(m1[k][j] == c1 && m2[i][k] == c1) {
						transformationMatrix[i][j] = c1;
						break;
					}
					else {
						transformationMatrix[i][j] = c0;
					}
				}
			}
		}
	}
}
